﻿/// <summary>
/// source: Http://www.WayneCliffordBarker.co.za
/// license: free to use by any individual or company for what ever purpose.
/// disclaimer: All content provided on this site "www.waynecliffordbarker.co.za" is for informational purposes only. Although the owner takes care to represent accurate content, the owner makes no representations as to the accuracy or completeness of any information on this site or found by following any link on this site. Neither the  owner of www.waynecliffordbarker.co.za nor any author of the site will be liable for any errors or omissions in this information nor for the availability of this information. The owner of www.waynecliffordbarker.co.za nor any author of the site will not be liable for any losses, injuries, or damages from the display or use of this information obtained from this site.
/// </summary>

using System;
using System.Collections.Generic;
using WayneCliffordBarker.RoutingEngine.Concerete;
using WayneCliffordBarker.RoutingEngine.Interface;
using WayneCliffordBarker.RoutingEngine.Model;

namespace WayneCliffordBarker.RoutingEngine
{
    /// <summary>
    /// Generic class to assist individuals that does validation, transformation and routing of data from one location to another.
    /// </summary>
    public sealed class RoutingEngine
    {

        public RoutingEngine(Ilogger Logger, IEnumerable<RoutingRule> Rules)
        {
            this.Rules = Rules;
            this.Logger = Logger;

        }

        private DataPacket dataPacket { get; set; }

        private IEnumerable<RoutingRule> Rules { get; set; }

        private Ilogger Logger { get; set; }
        
        private RoutingRule IdentifyRule()
        {
            this?.Logger?.WriteInformation("Identify rule");
            RoutingRule identifiedRule = null;

            foreach (RoutingRule rule in Rules)
            {
                rule.IdentifierInstance.Logger = this.Logger;
                if (rule.IdentifierInstance.Identify(dataPacket))
                {
                    this?.Logger?.WriteInformation("");
                    this?.Logger?.WriteAlert("Rule: " + rule.RuleDescription);
                    identifiedRule = rule;
                    break;
                }
            }

            if (identifiedRule == null)
            {
                throw new DataIdentificationException( "unable to identify rule");
            }

            return identifiedRule;
        }

        private void VerifyData(RoutingRule rule)
        {
            this?.Logger?.WriteInformation("Verify data");
            rule.ValidatorInstance.Logger = this.Logger;
            rule.ValidatorInstance.Validate(dataPacket);
        }

        private void ExecuteRule(RoutingRule rule)
        {
            this?.Logger?.WriteInformation("Executing rules");
            foreach (RoutingDestination destination in rule.Destinations)
            {                
                destination.TransformerInstance.Logger = this.Logger;
                IEnumerable<string> transformedData = destination.TransformerInstance.Transform(dataPacket.Content);

                this?.Logger?.WriteAlert("Persist data using " + destination.DeliveryMechanismInstance.GetType().ToString());
                destination.DeliveryMechanismInstance.Logger = this.Logger;
                destination.DeliveryMechanismInstance.Persist(transformedData);
                this?.Logger?.WriteInformation("");
            }           
        }

        private void PersistData()
        {
            PersistToSQL persist = new PersistToSQL(dataPacket.ToString());
            if (!persist.Persist())
            {
                throw new PersistException("Unable to persist message to storage.");
            }            
        }

        /// <summary>
        /// Route the <paramref name="message"/> according to the <see cref="RoutingRule"/>s that has been defined
        /// </summary>
        /// <param name="message">Raw message that was received that needs to be validated, transformed and send to other destinations</param>
        /// <returns>Boolean indicating if the data was successfully processed.</returns>
        public bool Route(DataPacket dataPacket)
        {
            if (dataPacket == null) throw new ArgumentNullException("dataPacket");


            this?.Logger?.WriteInformation(".................................................................");
            this.dataPacket = dataPacket;

            PersistData();

            RoutingRule identifiedRule = IdentifyRule();

            VerifyData(identifiedRule);

            ExecuteRule(identifiedRule);

            this?.Logger?.WriteInformation("");

            return true;
        }

        /// <summary>
        /// Route the <paramref name="data"/> according to the <see cref="RoutingRule"/>s that has been defined
        /// </summary>
        /// <param name="data">Raw message that was received that needs to be validated, transformed and send to other destinations</param>
        /// <returns>Boolean indicating if the data was successfully processed.</returns>
        public IEnumerable<Tuple<bool, DataPacket>> Route(IEnumerable<DataPacket> messages)
        {
            List<Tuple<bool, DataPacket>> results = new List<Tuple<bool, DataPacket>>();

            foreach (DataPacket dataPacket in messages)
            {
                bool result = false;

                try
                {
                    result = Route(dataPacket);
                }
                catch (Exception ex)
                {
                    this?.Logger?.WriteError("Skipping : " + ex.Message);
                    result = false;
                }

                results.Add( new Tuple<bool, DataPacket>(result, dataPacket));
            }

            return results;
        }
    }
}
